@@ -197,8 +197,7 @@ static int aread(struct align_fd *afd, void *buf, int len)
 
 	if (!bsize || bsize > 4096 || len > 4096) {
 		if (!bsize)
-			fprintf(stderr, "WARNING - aread() called with "
-				"invalid block size\n");
+			fprintf(stderr, "WARNING - aread() called with invalid block size\n");
 		return -1;
 	}
 	b = ROUND_UP_PTR((char *)abuf, 4096);
@@ -230,8 +229,7 @@ static int awrite(struct align_fd *afd, void *buf, int len)
 	bsize = afd->blk_sz;
 	if (!bsize || bsize > 4096 || len > 4096) {
 		if (!bsize)
-			fprintf(stderr, "WARNING - awrite() called with "
-				"invalid block size\n");
+			fprintf(stderr, "WARNING - awrite() called with invalid block size\n");
 		return -1;
 	}
 	b = ROUND_UP_PTR((char *)abuf, 4096);
@@ -811,7 +809,7 @@ static int examine_badblocks_super1(struct supertype *st, int fd, char *devname)
 
 	size = __le32_to_cpu(sb->bblog_size)* 512;
 	if (posix_memalign((void**)&bbl, 4096, size) != 0) {
-		pr_err("%s could not allocate badblocks list\n", __func__);
+		pr_err("could not allocate badblocks list\n");
 		return 0;
 	}
 	offset = __le64_to_cpu(sb->super_offset) +
@@ -1245,12 +1243,8 @@ static int update_super1(struct supertype *st, struct mdinfo *info,
 		/* set data_size to device size less data_offset */
 		struct misc_dev_info *misc = (struct misc_dev_info*)
 			(st->sb + MAX_SB_SIZE + BM_SUPER_SIZE);
-		printf("Size was %llu\n", (unsigned long long)
-		       __le64_to_cpu(sb->data_size));
 		sb->data_size = __cpu_to_le64(
 			misc->device_size - __le64_to_cpu(sb->data_offset));
-		printf("Size is %llu\n", (unsigned long long)
-		       __le64_to_cpu(sb->data_size));
 	} else if (strcmp(update, "revert-reshape") == 0) {
 		rv = -2;
 		if (!(sb->feature_map & __cpu_to_le32(MD_FEATURE_RESHAPE_ACTIVE)))
@@ -1332,7 +1326,7 @@ static int init_super1(struct supertype *st, mdu_array_info_t *info,
 	int sbsize;
 
 	if (posix_memalign((void**)&sb, 4096, SUPER1_SIZE) != 0) {
-		pr_err("%s could not allocate superblock\n", __func__);
+		pr_err("could not allocate superblock\n");
 		return 0;
 	}
 	memset(sb, 0, SUPER1_SIZE);
@@ -1557,7 +1551,7 @@ static int write_init_super1(struct supertype *st)
 	unsigned long long data_offset;
 
 	for (di = st->info; di; di = di->next) {
-		if (di->disk.state == 1)
+		if (di->disk.state & (1 << MD_DISK_FAULTY))
 			continue;
 		if (di->fd < 0)
 			continue;
@@ -1682,12 +1676,15 @@ static int write_init_super1(struct supertype *st)
 			}
 			break;
 		default:
-			pr_err("Failed to write invalid "
-			       "metadata format 1.%i to %s\n",
+			pr_err("Failed to write invalid metadata format 1.%i to %s\n",
 			       st->minor_version, di->devname);
 			rv = -EINVAL;
 			goto out;
 		}
+		if (conf_get_create_info()->bblist == 0) {
+			sb->bblog_size = 0;
+			sb->bblog_offset = 0;
+		}
 
 		sb->sb_csum = calc_sb_1_csum(sb);
 		rv = store_super1(st, di->fd);
@@ -1726,7 +1723,7 @@ static int compare_super1(struct supertype *st, struct supertype *tst)
 
 	if (!first) {
 		if (posix_memalign((void**)&first, 4096, SUPER1_SIZE) != 0) {
-			pr_err("%s could not allocate superblock\n", __func__);
+			pr_err("could not allocate superblock\n");
 			return 1;
 		}
 		memcpy(first, second, SUPER1_SIZE);
@@ -1837,8 +1834,7 @@ static int load_super1(struct supertype *st, int fd, char *devname)
 	}
 
 	if (posix_memalign((void**)&super, 4096, SUPER1_SIZE) != 0) {
-		pr_err("%s could not allocate superblock\n",
-			__func__);
+		pr_err("could not allocate superblock\n");
 		return 1;
 	}
 
@@ -2044,8 +2040,8 @@ add_internal_bitmap1(struct supertype *st,
 			 * been left.
 			 */
 			offset = 0;
-			room = choose_bm_space(__le64_to_cpu(sb->size));
 			bbl_size = 8;
+			room = choose_bm_space(__le64_to_cpu(sb->size)) + bbl_size;
 		} else {
 			room = __le64_to_cpu(sb->super_offset)
 				- __le64_to_cpu(sb->data_offset)
@@ -2071,8 +2067,8 @@ add_internal_bitmap1(struct supertype *st,
 	case 2: /* between superblock and data */
 		if (creating) {
 			offset = 4*2;
-			room = choose_bm_space(__le64_to_cpu(sb->size));
 			bbl_size = 8;
+			room = choose_bm_space(__le64_to_cpu(sb->size)) + bbl_size;
 		} else {
 			room = __le64_to_cpu(sb->data_offset)
 				- __le64_to_cpu(sb->super_offset);
@@ -2101,6 +2097,10 @@ add_internal_bitmap1(struct supertype *st,
 		/* Limit to 128K of bitmap when chunk size not requested */
 		room = 128*2;
 
+	if (room <= 1)
+		/* No room for a bitmap */
+		return 0;
+
 	max_bits = (room * 512 - sizeof(bitmap_super_t)) * 8;
 
 	min_chunk = 4096; /* sub-page chunks don't work yet.. */
